// eigen_min.h 2016-03-22 	minimal eigenvalue of symmetrix matrix
// (c) Khashin S.I. http://math.ivanovo.ac.ru/dalgebra/Khashin/index.html
// khash2 (at) gmail.com
//
#pragma warning(disable : 4996)

//#include <stdio.h>
#include <math.h>

#include "eigen_min.h"		// minimal eigenvalue of symmetrix matrix


static double vdbl_len2(int N, double *a)			// |a|
{
	double res = 0;
	for (int i = N; i>0; i--)
	{
		double d1 = (*a++);
		res += d1*d1;
	}
	return res;
}
static void vdbl_normal(int N, double *a)					// length(a):=1
{
	double len = sqrt(vdbl_len2(N,a));
	if (len == 0) return;
	len = 1 / len;
	for (int i = N; i>0; i--)
		(*a++) *= len;
}

static double vdbl_scprod(int N, double *a, double *b)			// (a,b)
{
	double res = 0;
	for (int i = N; i>0; i--)
		res += (*a++)*(*b++);
	return res;
}
static void vdbl_add(int N, double *a, double *b, double R)	// a += R*b
{
	for (int i = N; i>0; i--)
		(*a++) += R*(*b++);
}

static void vdbl_mult(int N, double *res, double *matr, double* v)	// res = matr*v
{
	for (int i = N; i>0; i--, matr += N)
		(*res++) = vdbl_scprod(N, matr, v);
}

#define AIJ(y,x) a[(y)*N+x]
#define DBL_MAXX 1.7976931348623158e+308

double  EigenMin(int N, double *a, double *v)		// v - eigenvectors of positive symmetric matrix with minimal eigenvalue (returned)
{
	int NN = N*N;
	int x, y, i;
	
	double eps = 0;
	for (i = 0; i < NN; i++){
		double t = a[i];
		eps += t*t;
	}
	eps = sqrt(eps)*1.0e-12;							// eps - almost 0!

														// check symmetry
	for (y = 1; y < N; y++)
		for (x = 0; x < y; x++)
			if (fabs(AIJ(x, y) - AIJ(y, x))>eps) {
				for (i = 0; i < N; i++) v[i] = 0;
				return DBL_MAXX;
			}

	// now matrix is symmetric
	double *Av  = new double[N];
	double *AAv = new double[N];

	for (x = 0; x < N; x++) v[x] = (1.11 + x) / (1 + x*x);		// some random values to vector v

	double f0, f1, f2, f3, res, a0, a1, a2, di, t1, t2, val1, val2;

	// main cycle
	for (i = 0; i < 3 * N; i++)
	{
		vdbl_normal(N, v);
		vdbl_mult(N, Av , a, v );
		vdbl_mult(N, AAv, a, Av);
		f0 = 1;
		f1 = vdbl_scprod(N, v,  Av );
		f2 = vdbl_scprod(N, v,  AAv);
		f3 = vdbl_scprod(N, Av, AAv);
		// function for minimize: f(t) = (f1 + 2*t*f2 + t^2*f3)/(f0 + 2*t*f1 + t^2*f2)
		res = f1 / f0;
		a0 = f1*f1 - f0*f2;
		a1 = f1*f2 - f3*f0;
		a2 = f2*f2 - f1*f3;
		if (fabs(a2) < eps)
			break;

		di = a1*a1 - 4 * a0*a2;			// must be >= 0!
		if (di < 0) 
			break;
		di = sqrt(di);
		if (di < eps)
			break;
		t1 = (-a1 + di) / (2 * a2);
		t2 = (-a1 - di) / (2 * a2);

		val1 = (f1 + 2 * t1*f2 + t1*t1 * f3) / (f0 + 2 * t1*f1 + t1*t1 * f2);
		val2 = (f1 + 2 * t2*f2 + t2*t2 * f3) / (f0 + 2 * t2*f1 + t2*t2 * f2);
		if (val1 <= val2) {
			vdbl_add(N, v, Av, t1);					// v += t1*Av;
			res = val1;
			if (fabs(t1) < 1e-11) 
				break;
		}
		else {
			vdbl_add(N, v, Av, t2);					// v += t1*Av;
			res = val2;
			if (fabs(t2) < 1e-11) 
				break;
		}
	}
	vdbl_normal(N, v);
	delete[]AAv;
	delete[]Av;
	return res;
}



